#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import logging

import pytest

from snowflake.core._utils import check_version_lte
from snowflake.core.database import Database
from snowflake.core.schema import Schema
from snowflake.core.version import __version__
from tests.utils import is_prod_or_preprod, random_string


@pytest.mark.min_sf_ver("99.99.99")
def test_should_never_run_in_prod_or_preprod(snowflake_version, snowflake_region):
    # This might still run in dev (where the version contains non-numerals,
    # so check if it has non-numerals). If it does not, then this should never
    # run.
    if is_prod_or_preprod(snowflake_version, snowflake_region):
        pytest.fail("This test should not have run in a prod or preprod env.")


@pytest.mark.min_sf_ver("1.0.0")
def test_should_always_run():
    pass


@pytest.mark.internal_only
@pytest.mark.usefixtures("backup_database_schema")
def test_large_results(databases, set_params):
    # Create a new db because it would only have 2 schemas initially: information_schema and public,
    # which does not trigger large results in the first iteration
    new_db = Database(name=random_string(3, "test_database_$12create_"), kind="TRANSIENT")
    database = databases.create(new_db)
    try:
        # This is fetched without large results
        schema_list1 = sorted(list(map(lambda sch: sch.name, database.schemas.iter())))

        with set_params(parameters={"RESULT_FIRST_CHUNK_MAX_SIZE": 1}, scope="session"):
            # This will be fetched with large results because we force the first chunk size to be small.
            schema_list2 = sorted(list(map(lambda sch: sch.name, database.schemas.iter())))
            assert schema_list1 == schema_list2
    finally:
        database.drop()


@pytest.mark.usefixtures("backup_database_schema")
def test_url_embedding_into_url(schemas, caplog):
    """Test whether URL part embedding works before logging.

    SNOW-1620036

    In the past we logged the URL we were reaching out to before all the paths
    were inserted. Leading to log lines like: "performing a HTTP POST call to
    /api/v2/databases/{database}/schemas". In this test we verify this does not
    happen anymore.
    """
    # We use schema because it's one of the top level objects that have db above
    new_schema = Schema(random_string(5, "test_url_embedding_into_url_"))
    with caplog.at_level(logging.INFO, logger="snowflake.core._generated.api_client"):
        s = schemas.create(new_schema)
    assert "{database}" not in caplog.text
    assert (f"performing a HTTP POST call to /api/v2/databases/{schemas.database.name}/schemas\n") in caplog.text
    caplog.clear()
    with caplog.at_level(logging.INFO, logger="snowflake.core._generated.api_client"):
        s.drop()
    assert "{database}" not in caplog.text
    assert (
        f"performing a HTTP DELETE call to /api/v2/databases/{schemas.database.name}/schemas/{new_schema.name}\n"
    ) in caplog.text


# Since our CI and tests are very unlikely to be a version that is not supported,
# this test should always pass -- we would never set the configuration to make it not work
# with the most recent version.


def test_client_info_integration(root):
    assert root._client_info.client_version == __version__

    # Assert that ClientInfo parsed the information correctly.
    assert root._client_info.minimum_supported_version is not None
    assert root._client_info.end_of_support_version is not None
    assert root._client_info.recommended_version is not None

    # All minimum and recommended versions should be less than or equal to the current version
    assert check_version_lte(root._client_info.minimum_supported_version, __version__)
    assert check_version_lte(root._client_info.end_of_support_version, __version__)
    assert check_version_lte(root._client_info.recommended_version, __version__)



from collections.abc import Iterator
from contextlib import suppress
from io import BytesIO
from textwrap import dedent

import pytest

import snowflake.connector

from snowflake.core import Root
from snowflake.core.compute_pool import ComputePool
from snowflake.core.database import Database, DatabaseCollection, DatabaseResource
from snowflake.core.exceptions import NotFoundError
from snowflake.core.image_repository import ImageRepository
from snowflake.core.schema import Schema, SchemaResource
from snowflake.core.service import Service, ServiceResource, ServiceSpecInlineText, ServiceSpecStageFile
from snowflake.core.stage import Stage, StageDirectoryTable, StageEncryption, StageResource
from snowflake.core.table import Table, TableColumn, TableResource
from snowflake.core.tag import Tag, TagResource

from ..utils import backup_role, random_string
from .constants import TEST_SCHEMA


@pytest.fixture(scope="session", autouse=True)
def test_schema() -> str:
    """Set up and tear down the test schema. This is automatically called per test session."""
    return TEST_SCHEMA


@pytest.fixture
def temp_ir(image_repositories) -> Iterator[ImageRepository]:
    ir_name = random_string(5, "test_ir_")
    test_ir = ImageRepository(
        name=ir_name
        # TODO: comment is not supported by image repositories?
        # comment="created by temp_ir",
    )
    image_repositories.create(test_ir)
    try:
        yield test_ir
    finally:
        image_repositories[test_ir.name].drop()


@pytest.fixture
def temp_cp(compute_pools, instance_family) -> Iterator[ComputePool]:
    cp_name = random_string(5, "test_cp_")
    test_cp = ComputePool(
        name=cp_name,
        instance_family=instance_family,
        min_nodes=1,
        max_nodes=1,
        auto_resume=False,
        comment="created by temp_cp",
    )
    compute_pools.create(test_cp, initially_suspended=True)
    try:
        yield test_cp
    finally:
        compute_pools[test_cp.name].drop()


@pytest.fixture
def temp_service(root, services, session, imagerepo, shared_compute_pool) -> Iterator[ServiceResource]:
    stage_name = random_string(5, "test_stage_")
    s_name = random_string(5, "test_service_")
    session.sql(f"create temp stage {stage_name};").collect()
    spec_file = "spec.yaml"
    spec = f"@{stage_name}/{spec_file}"
    session.file.put_stream(
        BytesIO(
            dedent(
                f"""
                spec:
                  containers:
                  - name: hello-world
                    image: {imagerepo}/hello-world:latest
                  endpoints:
                  - name: default
                    port: 8080
                 """
            ).encode()
        ),
        spec,
    )
    test_s = Service(
        name=s_name,
        compute_pool=shared_compute_pool,
        spec=ServiceSpecStageFile(stage=stage_name, spec_file=spec_file),
        min_instances=1,
        max_instances=1,
        comment="created by temp_service",
    )
    s = services.create(test_s)

    try:
        yield s
    finally:
        s.drop()


@pytest.fixture
def temp_service_from_spec_inline(root, services, session, imagerepo, shared_compute_pool) -> Iterator[ServiceResource]:
    s_name = random_string(5, "test_service_")
    inline_spec = dedent(
        f"""
        spec:
          containers:
          - name: hello-world
            image: {imagerepo}/hello-world:latest
         """
    )
    test_s = Service(
        name=s_name,
        compute_pool=shared_compute_pool,
        spec=ServiceSpecInlineText(spec_text=inline_spec),
        min_instances=1,
        max_instances=1,
        comment="created by temp_service_from_spec_inline",
    )
    s = services.create(test_s)
    try:
        yield test_s
    finally:
        s.drop()


@pytest.fixture
def temp_db(databases: DatabaseCollection, backup_database_schema) -> Iterator[DatabaseResource]:
    del backup_database_schema
    # create temp database
    db_name = random_string(5, "test_database_")
    test_db = Database(name=db_name, comment="created by temp_db")
    db = databases.create(test_db)
    try:
        yield db
    finally:
        db.drop()


@pytest.fixture
def temp_db_case_sensitive(databases: DatabaseCollection, backup_database_schema) -> Iterator[DatabaseResource]:
    del backup_database_schema
    # create temp database
    db_name = random_string(5, "test_database_case_sensitive_")
    db_name_case_sensitive = '"' + db_name + '"'
    test_db = Database(name=db_name_case_sensitive, comment="created by temp_case_sensitive_db")
    db = databases.create(test_db)
    try:
        yield db
    finally:
        db.drop()


@pytest.fixture
def temp_schema(schemas, backup_database_schema) -> SchemaResource:
    del backup_database_schema
    schema_name = random_string(5, "test_schema_")
    test_schema = Schema(name=schema_name, comment="created by temp_schema")
    sc = schemas.create(test_schema)
    try:
        yield sc
    finally:
        sc.drop()


@pytest.fixture
def temp_schema_case_sensitive(schemas, backup_database_schema) -> Iterator[SchemaResource]:
    del backup_database_schema
    schema_name = random_string(5, "test_schema_case_sensitive_")
    schema_name_case_sensitive = '"' + schema_name + '"'
    test_schema = Schema(name=schema_name_case_sensitive, comment="created by temp_schema_case_sensitive")
    sc = schemas.create(test_schema)
    try:
        yield sc
    finally:
        sc.drop()


@pytest.fixture
def temp_schema_upper_case(schemas, backup_database_schema) -> Iterator[SchemaResource]:
    del backup_database_schema
    schema_name_upper_case = random_string(5, "test_schema_upper_case_").upper()
    test_schema = Schema(name=schema_name_upper_case, comment="created by temp_schema_upper_case")
    sc = schemas.create(test_schema)
    try:
        yield sc
    finally:
        sc.drop()


@pytest.fixture(scope="session")
def temp_customer_organization(require_sf, sf_cursor) -> str:
    org_name = random_string(5, "testorg")
    with backup_role(sf_cursor):
        sf_cursor.execute("use role accountadmin;")
        sf_cursor.execute(
            f"CREATE ORGANIZATION {org_name} SALESFORCE_ID='new_salesforce_id' "
            "REGION_GROUPS=PUBLIC ORGANIZATION_TYPE='customer';"
        )
        try:
            yield org_name
        finally:
            sf_cursor.execute(f"DROP ORGANIZATION {org_name};")


@pytest.fixture(scope="session")
def temp_customer_account_with_org_admin(require_sf, temp_customer_organization, sf_cursor) -> str:
    with backup_role(sf_cursor):
        sf_cursor.execute("use role accountadmin;")
        sf_cursor.execute("alter session set QA_MODE=true;")
        account_name = random_string(5, "test_orgadmin_account_")
        blank_account_name = random_string(5, "test_blank_account_")
        sf_cursor.execute(f"CREATE ACCOUNT {blank_account_name} server_type=standard type=blank;")
        sf_cursor.execute(
            f"CREATE ACCOUNT {account_name} admin_name=admin admin_password='TestPassword1' "
            f"type=customer must_change_password=false organization={temp_customer_organization};"
        )
        sf_cursor.execute(f"call system$add_org_admin_to_account('{account_name.upper()}');")
        sf_cursor.execute(f"call system$it('GRANT_ORGADMIN_TO_ACCOUNTADMIN', '{account_name.upper()}');")
        sf_cursor.execute("alter session unset QA_MODE;")
        try:
            yield account_name
        finally:
            sf_cursor.execute(f"DROP ACCOUNT {account_name.upper()};")


@pytest.fixture(scope="session")
def temp_customer_account_root(temp_customer_account_with_org_admin) -> Root:
    with snowflake.connector.connect(
        **{
            "account": f"{temp_customer_account_with_org_admin}",
            "user": "admin",
            "password": "TestPassword1",
            "protocol": "http",
            "port": 8082,
            "host": f"{temp_customer_account_with_org_admin}.reg.local",
        }
    ) as connection:
        yield Root(connection)


@pytest.fixture(scope="session")
def temp_customer_account_cursor(temp_customer_account_root):
    return temp_customer_account_root._connection.cursor()


@pytest.fixture
def temp_stage(stages) -> Iterator[StageResource]:
    stage_name = random_string(5, "test_stage_")
    test_stage = Stage(name=stage_name, comment="created by temp_stage")
    st = stages.create(test_stage)
    try:
        yield st
    finally:
        st.drop()


@pytest.fixture
def temp_stage_case_sensitive(stages) -> Iterator[StageResource]:
    stage_name = random_string(5, "test_stage_case_sensitive_")
    stage_name_case_sensitive = '"' + stage_name + '"'
    test_stage = Stage(name=stage_name_case_sensitive, comment="created by temp_stage")
    st = stages.create(test_stage)
    try:
        yield st
    finally:
        st.drop()


@pytest.fixture
def temp_table(tables) -> Iterator[TableResource]:
    table_name = random_string(5, "test_table_case_insensitive_")
    columns = [TableColumn(name="col1", datatype="int"), TableColumn(name="col2", datatype="string")]
    test_table = Table(name=table_name, columns=columns)
    test_table_handle = tables.create(test_table)
    try:
        yield test_table_handle
    finally:
        with suppress(NotFoundError):
            test_table_handle.drop()


@pytest.fixture
def temp_directory_table(stages) -> Iterator[StageResource]:
    new_stage = Stage(
        name=random_string(5, "test_directory_table_"),
        encryption=StageEncryption(type="SNOWFLAKE_SSE"),
        directory_table=StageDirectoryTable(enable=True),
    )
    st = stages.create(new_stage)

    try:
        yield st
    finally:
        st.drop()


@pytest.fixture
def temp_tag_upper_case(temp_schema_upper_case) -> Iterator[TagResource]:
    tag_name = random_string(5, "temp_tag_").upper()
    temp_tag = temp_schema_upper_case.tags.create(Tag(name=tag_name))
    try:
        yield temp_schema_upper_case.tags[temp_tag.name]
    finally:
        temp_tag.drop(if_exists=True)



"""Manages Snowflake Notification Integrations."""

from ._generated.models import (
    NotificationEmail,
    NotificationHook,
    NotificationIntegration,
    NotificationQueueAwsSnsOutbound,
    NotificationQueueAzureEventGridInbound,
    NotificationQueueAzureEventGridOutbound,
    NotificationQueueGcpPubsubInbound,
    NotificationQueueGcpPubsubOutbound,
    NotificationWebhook,
    WebhookSecret,
)
from ._notification_integration import (
    NotificationIntegrationCollection,
    NotificationIntegrationResource,
)


__all__ = [
    "NotificationHook",
    "NotificationEmail",
    "NotificationWebhook",
    "NotificationQueueAwsSnsOutbound",
    "NotificationQueueAzureEventGridOutbound",
    "NotificationQueueGcpPubsubOutbound",
    "NotificationQueueAzureEventGridInbound",
    "NotificationQueueGcpPubsubInbound",
    "NotificationIntegration",
    "NotificationIntegrationCollection",
    "NotificationIntegrationResource",
    "WebhookSecret",
]



import pytest

from snowflake.core import CreateMode
from snowflake.core.database import Database
from snowflake.core.database_role import DatabaseRole
from snowflake.core.exceptions import ConflictError
from tests.integ.utils import backup_database_and_schema
from tests.utils import random_string


pytestmark = pytest.mark.min_sf_ver("8.39.0")


def test_create(databases, cursor):
    try:
        with backup_database_and_schema(cursor):
            new_db_def = Database(name=random_string(3, "test_database_$12create_"), kind="TRANSIENT")
            new_db_def.comment = "database first"
            database_handle = databases.create(new_db_def)
            cursor.execute(f"USE DATABASE {new_db_def.name}")

            database_role_name = random_string(10, "test_database_role")
            database_role = DatabaseRole(name=database_role_name, comment="test_comment")

            database_role_handles = []

            database_roles = databases[new_db_def.name].database_roles

            try:
                # check there are no database roles yet
                assert len(list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))) == 0

                # happy path
                database_role_handles += [database_roles.create(database_role)]
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 1
                assert database_roles_list[0].name == database_role_name.upper()
                assert database_roles_list[0].comment == "test_comment"

                # create role with already existing name with mode or_replace change comment
                database_role.comment = "test_comment_2"
                database_roles.create(database_role, mode=CreateMode.or_replace)
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 1
                assert database_roles_list[0].name == database_role_name.upper()
                assert database_roles_list[0].comment == "test_comment_2"

                # create role with already existing name with mode if_not_exists
                database_role.comment = "test_comment_3"
                database_roles.create(database_role, mode=CreateMode.if_not_exists)
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 1
                assert database_roles_list[0].name == database_role_name.upper()
                assert database_roles_list[0].comment == "test_comment_2"

                # create role with already existing name with mode error_if_exists
                database_role.comment = "test_comment_4"
                with pytest.raises(ConflictError):
                    database_roles.create(database_role, mode=CreateMode.error_if_exists)

            finally:
                for database_role_handle in database_role_handles:
                    database_role_handle.drop()
    finally:
        database_handle.drop()



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.tag import TagValue
from snowflake.core.task import Task, TaskResource
from snowflake.core.task._generated import TagAssignment, TagReference

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


@pytest.fixture
def tasks(schema):
    return schema.tasks


@pytest.fixture
def task(tasks):
    return tasks["my_task"]


def test_create_or_alter_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.create_or_alter_async(Task("my_task", "select 1"))
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root,
        "PUT",
        BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task",
        **extra_params(body={"name": "my_task", "definition": "select 1"}),
    )


def test_drop_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task", **extra_params()
    )


def test_fetch_async(fake_root, task):
    from snowflake.core.task._generated.models import Task as TaskModel

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(TaskModel(name="my_task", definition="select 1").to_json())
        op = task.fetch_async()
        assert isinstance(op, PollingOperation)
        task = op.result()
        assert task.to_dict() == Task(name="my_task", definition="select 1").to_dict()
    mocked_request.assert_called_once_with(
        fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task", **extra_params()
    )


def test_execute_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.execute_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task:execute?retryLast=False",
        **extra_params(query_params=[("retryLast", False)]),
    )


def test_resume_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.resume_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task:resume", **extra_params()
    )


def test_suspend_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.suspend_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task:suspend", **extra_params()
    )


def test_fetch_task_dependents_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = task.fetch_task_dependents_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task/dependents", **extra_params()
    )


def test_get_complete_graphs_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.get_complete_graphs_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root,
        "GET",
        BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task/complete-graphs?errorOnly=True",
        **extra_params(query_params=[("errorOnly", True)]),
    )


def test_get_current_graphs_async(fake_root, task):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.get_current_graphs_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(
        fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task/current-graphs", **extra_params()
    )


def test_create_async(fake_root, tasks):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = tasks.create_async(Task("my_task", "select 1"))
        assert isinstance(op, PollingOperation)
        task_res = op.result()
        assert isinstance(task_res, TaskResource)
        assert task_res.name == "my_task"
    mocked_request.assert_called_once_with(
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/tasks?createMode=errorIfExists",
        **extra_params(
            query_params=[("createMode", "errorIfExists")], body={"name": "my_task", "definition": "select 1"}
        ),
    )


def test_iter_async(fake_root, tasks):
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = tasks.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(
        fake_root,
        "GET",
        BASE_URL + "/databases/my_db/schemas/my_schema/tasks?rootOnly=False",
        **extra_params(query_params=[("rootOnly", False)]),
    )


def test_set_tags(fake_root, task, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        task.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, task, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        task.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = task.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, task):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/tasks/my_task:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert task.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = task.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



import pytest

from snowflake.core.role import ContainingScope, Securable
from snowflake.core.table import Table, TableColumn
from tests.integ.role.utils import assert_basic_grant
from tests.utils import random_string


pytestmark = pytest.mark.min_sf_ver("8.39.0")


@pytest.mark.use_accountadmin
def test_iter_grant_to(
    roles,
    test_role_name,
    test_role_name_2,
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    test_table_for_grant_name,
    tables_in_test_database,
):
    # test grant role
    roles[test_role_name].grant_role(role_type="ROLE", role=Securable(name=test_role_name_2))
    grants_list = list(roles[test_role_name].iter_grants_to())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert_basic_grant(grant)
    assert not grant.grant_option
    assert grant.securable.name == test_role_name_2.upper()
    assert grant.securable_type == "ROLE"

    roles[test_role_name].revoke_role(role_type="ROLE", role=Securable(name=test_role_name_2))

    # test grant database role becasue in this we will get the securable database name
    roles[test_role_name].grant_role(
        role_type="DATABASE ROLE", role=Securable(name=test_database_role_name, database=test_database_for_grant_name)
    )
    grants_list = list(roles[test_role_name].iter_grants_to())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert_basic_grant(grant)
    assert not grant.grant_option
    assert grant.securable.name == test_database_role_name.upper()
    assert grant.securable.database == test_database_for_grant_name.upper()
    assert grant.securable_type == "DATABASE ROLE"

    roles[test_role_name].revoke_role(
        role_type="DATABASE ROLE", role=Securable(name=test_database_role_name, database=test_database_for_grant_name)
    )

    # without securable name (to this account)
    roles[test_role_name].grant_privileges(privileges=["CREATE DATABASE"], securable_type="ACCOUNT")
    grants_list = list(roles[test_role_name].iter_grants_to())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert_basic_grant(grant)
    assert grant.securable.name is not None
    assert grant.securable_type == "ACCOUNT"
    assert grant.securable.database is None
    assert grant.securable.var_schema is None

    roles[test_role_name].revoke_privileges(privileges=["CREATE DATABASE"], securable_type="ACCOUNT")

    # grant on a schema here we will get the securable name as well as schema name
    roles[test_role_name].grant_privileges(
        privileges=["CREATE TASK"],
        securable_type="SCHEMA",
        securable=Securable(database=test_database_for_grant_name, name=test_schema_for_grant_name),
    )
    grants_list = list(roles[test_role_name].iter_grants_to())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert_basic_grant(grant)
    assert grant.securable.name == test_schema_for_grant_name.upper()
    assert grant.securable_type == "SCHEMA"
    assert grant.securable.database == test_database_for_grant_name

    roles[test_role_name].revoke_privileges(
        privileges=["CREATE TASK"],
        securable_type="SCHEMA",
        securable=Securable(database=test_database_for_grant_name, name=test_schema_for_grant_name),
    )

    # grant on a table
    roles[test_role_name].grant_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )
    grants_list = list(roles[test_role_name].iter_grants_to())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert_basic_grant(grant)
    assert grant.securable.name == test_table_for_grant_name.upper()
    assert grant.securable_type == "TABLE"
    assert grant.securable.database == test_database_for_grant_name.upper()
    assert grant.securable.var_schema == test_schema_for_grant_name.upper()

    roles[test_role_name].revoke_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    # all future tables in a schema
    roles[test_role_name].grant_future_privileges(
        privileges=["INSERT"],
        securable_type="TABLE",
        containing_scope=ContainingScope(database=test_database_for_grant_name, schema=test_schema_for_grant_name),
    )

    grants_list = list(roles[test_role_name].iter_grants_to())
    assert len(grants_list) == 0

    try:
        table_name = random_string(5, "test_grant_future_privileges_table_")
        columns = [TableColumn(name="col1", datatype="int"), TableColumn(name="col2", datatype="string")]
        test_table = Table(name=table_name, columns=columns)
        tables_in_test_database.create(test_table)

        grants_list = list(roles[test_role_name].iter_grants_to())
        assert len(grants_list) == 1
        grant = grants_list[0]
        assert_basic_grant(grant)
        assert grant.securable.name == table_name.upper()
        assert grant.securable_type == "TABLE"
        assert grant.securable.database == test_database_for_grant_name.upper()
        assert grant.securable.var_schema == test_schema_for_grant_name.upper()
    finally:
        tables_in_test_database[table_name].drop()


@pytest.mark.use_accountadmin
def test_iter_grant_on(roles, test_role_name):
    # because the role was created by accountadmin, accountadmin alreay have ownership privilages on the role
    grants_list = list(roles[test_role_name].iter_grants_on())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert grant.created_on is not None
    assert grant.privilege is not None
    assert grant.granted_on == "ROLE"
    assert grant.name == test_role_name.upper()
    assert grant.granted_to == "ROLE"
    assert grant.granted_by == "ACCOUNTADMIN"
    assert grant.granted_by_role_type == "ROLE"
    assert grant.grantee_name == "ACCOUNTADMIN"
    assert grant.grant_option


@pytest.mark.use_accountadmin
def test_iter_grant_of(roles, test_role_name, test_role_name_2):
    grants_list = list(roles[test_role_name].iter_grants_of())
    assert len(grants_list) == 0
    # test grant role
    roles[test_role_name_2].grant_role(role_type="ROLE", role=Securable(name=test_role_name))
    grants_list = list(roles[test_role_name].iter_grants_of())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert grant.created_on is not None
    assert grant.granted_by == "ACCOUNTADMIN"
    assert grant.grantee_name == test_role_name_2.upper()
    assert grant.granted_to == "ROLE"
    assert grant.role == test_role_name.upper()


@pytest.mark.use_accountadmin
def test_iter_future_grant_to(roles, test_role_name, test_database_for_grant_name, test_schema_for_grant_name):
    assert len(list(roles[test_role_name].iter_future_grants_to())) == 0
    # all future tables in a schema
    roles[test_role_name].grant_future_privileges(
        privileges=["INSERT"],
        securable_type="TABLE",
        containing_scope=ContainingScope(database=test_database_for_grant_name, schema=test_schema_for_grant_name),
    )

    grants_list = list(roles[test_role_name].iter_future_grants_to())
    assert len(grants_list) == 1
    grant = grants_list[0]
    assert_basic_grant(grant, False)  # future grants do not have granted_by
    assert grant.securable.name == '"<TABLE>"'
    assert grant.securable_type == "TABLE"
    assert grant.securable.database == test_database_for_grant_name.upper()
    assert grant.securable.var_schema == test_schema_for_grant_name.upper()



from unittest import mock

import pytest

from snowflake.core import PollingOperation, Root
from snowflake.core.event_table import EventTable, EventTableResource
from snowflake.core.event_table._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


@pytest.fixture
def event_tables(schema):
    return schema.event_tables


@pytest.fixture
def event_table(event_tables):
    return event_tables["my_table"]


def test_create_event_table(fake_root, event_tables):
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/event-tables?createMode=errorIfExists&copyGrants=False",
    )
    kwargs = extra_params(
        query_params=[("createMode", "errorIfExists"), ("copyGrants", False)], body={"name": "my_table"}
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        et_res = event_tables.create(EventTable(name="my_table"))
        assert isinstance(et_res, EventTableResource)
        assert et_res.name == "my_table"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = event_tables.create_async(EventTable(name="my_table"))
        assert isinstance(op, PollingOperation)
        et_res = op.result()
        assert et_res.name == "my_table"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_event_table(fake_root, event_tables):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/event-tables")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        event_tables.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = event_tables.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_event_table(fake_root, event_table):
    from snowflake.core.event_table._generated.models import EventTable as EventTableModel

    model = EventTableModel(name="my_table")
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/event-tables/my_table")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        event_table.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = event_table.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == EventTable(name="my_table").to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_event_table(fake_root, event_table):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/event-tables/my_table")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        event_table.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = event_table.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_rename_event_table(fake_root, event_table, event_tables):
    def format_args(table_name: str) -> tuple[Root, str, str]:
        return (
            fake_root,
            "POST",
            BASE_URL + f"/databases/my_db/schemas/my_schema/event-tables/{table_name}:rename?targetName=new_table",
        )

    kwargs = extra_params(query_params=[("targetName", "new_table")])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        event_table.rename("new_table")
        assert event_table.name == "new_table"
    mocked_request.assert_called_once_with(*format_args("my_table"), **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        et_res = event_tables["another_table"]
        op = et_res.rename_async("new_table")
        assert isinstance(op, PollingOperation)
        op.result()
        assert et_res.name == "new_table"
    mocked_request.assert_called_once_with(*format_args("another_table"), **kwargs)


def test_set_tags(fake_root, event_table, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/event-tables/my_table:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        event_table.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = event_table.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, event_table, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/event-tables/my_table:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        event_table.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = event_table.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, event_table):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/event-tables/my_table:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert event_table.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = event_table.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
import pytest

from snowflake.core.image_repository import ImageRepository
from tests.utils import random_string


def test_iter(image_repositories):
    ir_names = []
    try:
        ir_name = random_string(5, "test_ir_1")
        ir_names += [ir_name]
        test_ir = ImageRepository(name=ir_name)
        image_repo_1 = image_repositories.create(test_ir)

        ir_name = random_string(5, "test_ir_2")
        ir_names += [ir_name]
        test_ir = ImageRepository(name=ir_name)
        image_repo_2 = image_repositories.create(test_ir)

        image_repo_names = [image_repo.name for image_repo in image_repositories.iter()]

        assert image_repo_1.name.upper() in image_repo_names
        assert image_repo_2.name.upper() in image_repo_names

        image_repo_names = [image_repo.name for image_repo in image_repositories.iter(like="TEst_Ir_%")]

        assert image_repo_1.name.upper() in image_repo_names
        assert image_repo_2.name.upper() in image_repo_names

        image_repo_names = [image_repo.name for image_repo in image_repositories.iter(like="TEst_Ir_2%")]

        assert image_repo_1.name.upper() not in image_repo_names
        assert image_repo_2.name.upper() in image_repo_names

        image_repo_names = [image_repo.name for image_repo in image_repositories.iter(like="TEst_Ir_3%")]

        assert image_repo_1.name.upper() not in image_repo_names
        assert image_repo_2.name.upper() not in image_repo_names

    finally:
        for ir_name in ir_names:
            image_repositories[ir_name].drop(if_exists=True)


@pytest.mark.skip_gov
def test_list_images(image_repositories):
    try:
        ir_name = random_string(5, "test_ir_3")
        test_ir = ImageRepository(name=ir_name)
        image_repo_1 = image_repositories.create(test_ir)

        images = image_repo_1.list_images_in_repository()

        assert len(list(images)) == 0
    finally:
        image_repo_1.drop(if_exists=True)



"""
Snowflake Pipe API.

The Snowflake Pipe API is a REST API that you can use to access, update, and perform certain actions on Pipe resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictStr, field_validator
from typing_extensions import Annotated


class Pipe(BaseModel):
    """A model object representing the Pipe resource.

    Constructs an object of type Pipe with the provided properties.

    Parameters
    __________
    name : str
        Name of the pipe
    copy_statement : str
        COPY INTO <table> statement used to load data from queued files into a Snowflake table. This statement serves as the text/definition for the pipe and is displayed in the SHOW PIPES output
    comment : str, optional
        user comment associated to an object in the dictionary
    auto_ingest : bool, optional
        TRUE if all files from stage need to be auto-ingested
    error_integration : str, optional
        Link to integration object that point to a user provided Azure storage queue / SQS. When present,  errors (e.g. ingest failure for Snowpipe or a user task failure or replication failure) will be sent to this queue to notify customers
    aws_sns_topic : str, optional
        Optional, if provided, auto_ingest pipe will only receive messages from this SNS topic.
    integration : str, optional
        Link to integration object that ties a user provided storage queue to an auto_ingest enabled pipe. Required for auto_ingest to work on azure.
    created_on : datetime, optional
        Date and time when the pipe was created — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database in which the pipe is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the pipe is stored — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Role that owns the pipe — **Read-only:** *any user-provided value will be ignored.*
    pattern : str, optional
        PATTERN copy option value in the COPY INTO <table> statement in the pipe definition, if the copy option was specified — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        The type of role that owns the pipe — **Read-only:** *any user-provided value will be ignored.*
    invalid_reason : str, optional
        Displays some detailed information for your pipes that may have issues — **Read-only:** *any user-provided value will be ignored.*
    budget : str, optional
        Name of the budget if the pipe is monitored by a budget — **Read-only:** *any user-provided value will be ignored.*
    """

    name: StrictStr

    comment: Optional[StrictStr] = None

    auto_ingest: Optional[StrictBool] = None

    error_integration: Optional[StrictStr] = None

    aws_sns_topic: Optional[StrictStr] = None

    integration: Optional[StrictStr] = None

    copy_statement: Annotated[str, Field(strict=True)]

    created_on: Optional[datetime] = None

    database_name: Optional[StrictStr] = None

    schema_name: Optional[StrictStr] = None

    owner: Optional[StrictStr] = None

    pattern: Optional[StrictStr] = None

    owner_role_type: Optional[StrictStr] = None

    invalid_reason: Optional[StrictStr] = None

    budget: Optional[StrictStr] = None

    __properties = [
        "name",
        "comment",
        "auto_ingest",
        "error_integration",
        "aws_sns_topic",
        "integration",
        "copy_statement",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "pattern",
        "owner_role_type",
        "invalid_reason",
        "budget",
    ]

    @field_validator("copy_statement")
    def copy_statement_validate_regular_expression(cls, v):
        if not re.match(r"""(?i)^COPY INTO .*""", v):
            raise ValueError(r"""must validate the regular expression /(?i)^COPY INTO .*/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Pipe:
        """Create an instance of Pipe from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "owner",
                    "pattern",
                    "owner_role_type",
                    "invalid_reason",
                    "budget",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Pipe:
        """Create an instance of Pipe from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Pipe.model_validate(obj)

        _obj = Pipe.model_validate(
            {
                "name": obj.get("name"),
                "comment": obj.get("comment"),
                "auto_ingest": obj.get("auto_ingest"),
                "error_integration": obj.get("error_integration"),
                "aws_sns_topic": obj.get("aws_sns_topic"),
                "integration": obj.get("integration"),
                "copy_statement": obj.get("copy_statement"),
                "created_on": obj.get("created_on"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "owner": obj.get("owner"),
                "pattern": obj.get("pattern"),
                "owner_role_type": obj.get("owner_role_type"),
                "invalid_reason": obj.get("invalid_reason"),
                "budget": obj.get("budget"),
            }
        )

        return _obj


class PipeModel:
    def __init__(
        self,
        name: str,
        copy_statement: str,
        # optional properties
        comment: Optional[str] = None,
        auto_ingest: Optional[bool] = None,
        error_integration: Optional[str] = None,
        aws_sns_topic: Optional[str] = None,
        integration: Optional[str] = None,
        created_on: Optional[datetime] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        pattern: Optional[str] = None,
        owner_role_type: Optional[str] = None,
        invalid_reason: Optional[str] = None,
        budget: Optional[str] = None,
    ):
        """A model object representing the Pipe resource.

        Constructs an object of type Pipe with the provided properties.

        Parameters
        __________
        name : str
            Name of the pipe
        copy_statement : str
            COPY INTO <table> statement used to load data from queued files into a Snowflake table. This statement serves as the text/definition for the pipe and is displayed in the SHOW PIPES output
        comment : str, optional
            user comment associated to an object in the dictionary
        auto_ingest : bool, optional
            TRUE if all files from stage need to be auto-ingested
        error_integration : str, optional
            Link to integration object that point to a user provided Azure storage queue / SQS. When present,  errors (e.g. ingest failure for Snowpipe or a user task failure or replication failure) will be sent to this queue to notify customers
        aws_sns_topic : str, optional
            Optional, if provided, auto_ingest pipe will only receive messages from this SNS topic.
        integration : str, optional
            Link to integration object that ties a user provided storage queue to an auto_ingest enabled pipe. Required for auto_ingest to work on azure.
        created_on : datetime, optional
            Date and time when the pipe was created.
        database_name : str, optional
            Database in which the pipe is stored
        schema_name : str, optional
            Schema in which the pipe is stored
        owner : str, optional
            Role that owns the pipe
        pattern : str, optional
            PATTERN copy option value in the COPY INTO <table> statement in the pipe definition, if the copy option was specified.
        owner_role_type : str, optional
            The type of role that owns the pipe
        invalid_reason : str, optional
            Displays some detailed information for your pipes that may have issues
        budget : str, optional
            Name of the budget if the pipe is monitored by a budget
        """
        self.name = name
        self.comment = comment
        self.auto_ingest = auto_ingest
        self.error_integration = error_integration
        self.aws_sns_topic = aws_sns_topic
        self.integration = integration
        self.copy_statement = copy_statement
        self.created_on = created_on
        self.database_name = database_name
        self.schema_name = schema_name
        self.owner = owner
        self.pattern = pattern
        self.owner_role_type = owner_role_type
        self.invalid_reason = invalid_reason
        self.budget = budget

    __properties = [
        "name",
        "comment",
        "auto_ingest",
        "error_integration",
        "aws_sns_topic",
        "integration",
        "copy_statement",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "pattern",
        "owner_role_type",
        "invalid_reason",
        "budget",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Pipe(
            name=self.name,
            comment=self.comment,
            auto_ingest=self.auto_ingest,
            error_integration=self.error_integration,
            aws_sns_topic=self.aws_sns_topic,
            integration=self.integration,
            copy_statement=self.copy_statement,
            created_on=self.created_on,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            pattern=self.pattern,
            owner_role_type=self.owner_role_type,
            invalid_reason=self.invalid_reason,
            budget=self.budget,
        )

    @classmethod
    def _from_model(cls, model) -> PipeModel:
        return PipeModel(
            name=model.name,
            comment=model.comment,
            auto_ingest=model.auto_ingest,
            error_integration=model.error_integration,
            aws_sns_topic=model.aws_sns_topic,
            integration=model.integration,
            copy_statement=model.copy_statement,
            created_on=model.created_on,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            pattern=model.pattern,
            owner_role_type=model.owner_role_type,
            invalid_reason=model.invalid_reason,
            budget=model.budget,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Pipe.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Pipe object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> PipeModel:
        """Create an instance of Pipe from a dict.

        This method constructs a Pipe object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Pipe
            A Pipe object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Pipe.from_dict(obj))


Pipe._model_class = PipeModel



"""
Cortex Inference API.

OpenAPI 3.0 specification for the Cortex REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict

from snowflake.core.cortex.inference_service._generated.models.streaming_complete_response_data_event_choices_inner import (
    StreamingCompleteResponseDataEventChoicesInner,
    StreamingCompleteResponseDataEventChoicesInnerModel,
)


class StreamingCompleteResponseDataEvent(BaseModel):
    """A model object representing the StreamingCompleteResponseDataEvent resource.

    Constructs an object of type StreamingCompleteResponseDataEvent with the provided properties.

    Parameters
    __________
    choices : list[StreamingCompleteResponseDataEventChoicesInner], optional
    """

    choices: Optional[List[StreamingCompleteResponseDataEventChoicesInner]] = None

    __properties = ["choices"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> StreamingCompleteResponseDataEvent:
        """Create an instance of StreamingCompleteResponseDataEvent from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in choices (list)
        _items = []
        if self.choices:
            for _item in self.choices:
                if _item:
                    _items.append(_item.to_dict())
            _dict["choices"] = _items

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> StreamingCompleteResponseDataEvent:
        """Create an instance of StreamingCompleteResponseDataEvent from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return StreamingCompleteResponseDataEvent.model_validate(obj)

        _obj = StreamingCompleteResponseDataEvent.model_validate(
            {
                "choices": [
                    StreamingCompleteResponseDataEventChoicesInner.from_dict(_item) for _item in obj.get("choices")
                ]
                if obj.get("choices") is not None
                else None,
            }
        )

        return _obj


class StreamingCompleteResponseDataEventModel:
    def __init__(
        self,  # optional properties
        choices: Optional[list[StreamingCompleteResponseDataEventChoicesInner]] = None,
    ):
        """A model object representing the StreamingCompleteResponseDataEvent resource.

        Constructs an object of type StreamingCompleteResponseDataEvent with the provided properties.

        Parameters
        __________
        choices : list[StreamingCompleteResponseDataEventChoicesInner], optional
        """
        self.choices = choices

    __properties = ["choices"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return StreamingCompleteResponseDataEvent(
            choices=[x._to_model() for x in self.choices] if self.choices is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> StreamingCompleteResponseDataEventModel:
        return StreamingCompleteResponseDataEventModel(
            choices=[StreamingCompleteResponseDataEventChoicesInnerModel._from_model(x) for x in model.choices]
            if model.choices
            else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a StreamingCompleteResponseDataEvent.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the StreamingCompleteResponseDataEvent object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> StreamingCompleteResponseDataEventModel:
        """Create an instance of StreamingCompleteResponseDataEvent from a dict.

        This method constructs a StreamingCompleteResponseDataEvent object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        StreamingCompleteResponseDataEvent
            A StreamingCompleteResponseDataEvent object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(StreamingCompleteResponseDataEvent.from_dict(obj))


StreamingCompleteResponseDataEvent._model_class = StreamingCompleteResponseDataEventModel



"""
Cortex Embed API.

OpenAPI 3.0 specification for the Cortex Embed REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List

from pydantic import BaseModel, ConfigDict, StrictStr, field_validator

from snowflake.core.cortex.embed_service._generated.models.embed_response_data_inner import (
    EmbedResponseDataInner,
    EmbedResponseDataInnerModel,
)
from snowflake.core.cortex.embed_service._generated.models.embed_response_usage import (
    EmbedResponseUsage,
    EmbedResponseUsageModel,
)


class EmbedResponse(BaseModel):
    """A model object representing the EmbedResponse resource.

    Constructs an object of type EmbedResponse with the provided properties.

    Parameters
    __________
    object : str
        Type of the response object
    data : list[EmbedResponseDataInner]
        List of embedding objects
    model : str
        Name of the embedding model used
    usage : EmbedResponseUsage
    """

    object: StrictStr

    data: List[EmbedResponseDataInner]

    model: StrictStr

    usage: EmbedResponseUsage

    __properties = ["object", "data", "model", "usage"]

    @field_validator("object")
    def object_validate_enum(cls, v):
        if v not in ("list"):
            raise ValueError("must validate the enum values ('list')")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> EmbedResponse:
        """Create an instance of EmbedResponse from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in data (list)
        _items = []
        if self.data:
            for _item in self.data:
                if _item:
                    _items.append(_item.to_dict())
            _dict["data"] = _items

        # override the default output from pydantic by calling `to_dict()` of usage
        if self.usage:
            _dict["usage"] = self.usage.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> EmbedResponse:
        """Create an instance of EmbedResponse from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return EmbedResponse.model_validate(obj)

        _obj = EmbedResponse.model_validate(
            {
                "object": obj.get("object"),
                "data": [EmbedResponseDataInner.from_dict(_item) for _item in obj.get("data")]
                if obj.get("data") is not None
                else None,
                "model": obj.get("model"),
                "usage": EmbedResponseUsage.from_dict(obj.get("usage")) if obj.get("usage") is not None else None,
            }
        )

        return _obj


class EmbedResponseModel:
    def __init__(
        self,
        object: str,
        data: list[EmbedResponseDataInner],
        model: str,
        usage: EmbedResponseUsage,
        # optional properties
    ):
        """A model object representing the EmbedResponse resource.

        Constructs an object of type EmbedResponse with the provided properties.

        Parameters
        __________
        object : str
            Type of the response object
        data : list[EmbedResponseDataInner]
            List of embedding objects
        model : str
            Name of the embedding model used
        usage : EmbedResponseUsage
        """
        self.object = object
        self.data = data
        self.model = model
        self.usage = usage

    __properties = ["object", "data", "model", "usage"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return EmbedResponse(
            object=self.object,
            data=[x._to_model() for x in self.data] if self.data is not None else None,
            model=self.model,
            usage=self.usage._to_model() if self.usage is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> EmbedResponseModel:
        return EmbedResponseModel(
            object=model.object,
            data=[EmbedResponseDataInnerModel._from_model(x) for x in model.data] if model.data else None,
            model=model.model,
            usage=EmbedResponseUsageModel._from_model(model.usage) if model.usage else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a EmbedResponse.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the EmbedResponse object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> EmbedResponseModel:
        """Create an instance of EmbedResponse from a dict.

        This method constructs a EmbedResponse object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        EmbedResponse
            A EmbedResponse object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(EmbedResponse.from_dict(obj))


EmbedResponse._model_class = EmbedResponseModel



from contextlib import suppress

import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.user_defined_function import (
    Argument,
    ColumnType,
    JavaFunction,
    JavaScriptFunction,
    PythonFunction,
    ReturnDataType,
    ReturnTable,
    ScalaFunction,
    SQLFunction,
    UserDefinedFunction,
)
from tests.utils import random_string


pytestmark = [pytest.mark.usefixtures("anaconda_package_available")]


def test_create_and_fetch_python(user_defined_functions):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_py_")

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[],
            return_type=ReturnDataType(datatype="VARIANT"),
            language_config=PythonFunction(runtime_version="3.13", packages=[], handler="udf"),
            body="""
def udf():
    return {"key": "value"}
            """,
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "VARIANT"
        assert user_defined_function_handle.language_config.runtime_version == "3.13"
        assert user_defined_function_handle.language_config.packages == []
        assert user_defined_function_handle.language_config.handler == "udf"
        assert (
            user_defined_function_handle.body
            == """
def udf():
    return {"key": "value"}
            """
        )

    finally:
        user_defined_function_created.drop()


def test_create_and_fetch_java(user_defined_functions, cursor):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_java_")
    target_path = random_string(5, "@~/testfunc_") + ".jar"

    func_body = """
    class TestFunc {
        public static String echoVarchar(String x) {
            return x;
        }
    }
    """

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[Argument(name="x", datatype="STRING")],
            return_type=ReturnDataType(datatype="VARCHAR", nullable=True),
            language_config=JavaFunction(
                handler="TestFunc.echoVarchar",
                runtime_version="11",
                target_path=target_path,
                packages=[],
                called_on_null_input=True,
                is_volatile=True,
            ),
            body=func_body,
            comment="test_comment",
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "VARCHAR"
        assert user_defined_function_handle.return_type.nullable is True
        assert user_defined_function_handle.language_config.runtime_version == "11"
        assert user_defined_function_handle.language_config.packages == []
        assert user_defined_function_handle.language_config.handler == "TestFunc.echoVarchar"
        assert user_defined_function_handle.language_config.target_path == target_path
        assert user_defined_function_handle.body == func_body
        assert user_defined_function_handle.comment == "test_comment"

    finally:
        user_defined_function_created.drop()
        with suppress(NotFoundError):
            cursor.execute(f"RM @~ pattern='.*{target_path.split('/')[1]}.*'").fetchone()


def test_create_and_fetch_js(user_defined_functions):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_js_")

    func_body = """
        if (D <= 0) {
        return 1;
        } else {
        var result = 1;
        for (var i = 2; i <= D; i++) {
            result = result * i;
        }
        return result;
        }
    """

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[Argument(name="d", datatype="DOUBLE")],
            return_type=ReturnDataType(datatype="DOUBLE"),
            language_config=JavaScriptFunction(),
            body=func_body,
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "FLOAT"
        assert user_defined_function_handle.body == func_body

    finally:
        user_defined_function_created.drop()


def test_create_and_fetch_scala(user_defined_functions, cursor):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_scala_")
    target_path = random_string(5, "@~/testfunc_scala") + ".jar"

    func_body = """
        class Echo {
            def echoVarchar(x : String): String = {
                return x
            }
        }
    """

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[Argument(name="x", datatype="VARCHAR")],
            return_type=ReturnDataType(datatype="VARCHAR"),
            language_config=ScalaFunction(
                runtime_version="2.12", handler="Echo.echoVarchar", target_path=target_path, packages=[]
            ),
            body=func_body,
            comment="test_comment",
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "VARCHAR"
        assert user_defined_function_handle.language_config.runtime_version == "2.12"
        assert user_defined_function_handle.language_config.packages == []
        assert user_defined_function_handle.language_config.handler == "Echo.echoVarchar"
        assert user_defined_function_handle.language_config.target_path == target_path
        assert user_defined_function_handle.body == func_body
        assert user_defined_function_handle.comment == "test_comment"

    finally:
        user_defined_function_created.drop()

    user_defined_function_name = random_string(10, "test_create_user_defined_function_scala_staged_")
    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[Argument(name="x", datatype="VARCHAR")],
            return_type=ReturnDataType(datatype="VARCHAR"),
            language_config=ScalaFunction(
                runtime_version="2.12", handler="Echo.echoVarchar", imports=[target_path], packages=[]
            ),
            comment="test_comment",
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "VARCHAR"
        assert user_defined_function_handle.language_config.runtime_version == "2.12"
        assert user_defined_function_handle.language_config.packages == []
        assert user_defined_function_handle.language_config.handler == "Echo.echoVarchar"
        assert user_defined_function_handle.language_config.imports == [target_path]
        assert user_defined_function_handle.comment == "test_comment"

    finally:
        user_defined_function_created.drop()
        with suppress(NotFoundError):
            cursor.execute(f"RM @~ pattern='.*{target_path.split('/')[1]}.*'").fetchone()


def test_create_and_fetch_scala_object_argument(user_defined_functions):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_scala_object_arg_")

    func_body = """
        class VariantLibrary {
          def extract(m: Map[String, String], key: String): String = {
            return m(key)
          }
        }
    """

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[Argument(name="x", datatype="OBJECT"), Argument(name="key", datatype="VARCHAR")],
            return_type=ReturnDataType(datatype="VARIANT"),
            language_config=ScalaFunction(handler="VariantLibrary.extract", runtime_version="2.12", packages=[]),
            body=func_body,
            comment="test_comment",
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "VARIANT"
        assert user_defined_function_handle.language_config.runtime_version == "2.12"
        assert user_defined_function_handle.language_config.packages == []
        assert user_defined_function_handle.language_config.handler == "VariantLibrary.extract"
        assert user_defined_function_handle.body == func_body
        assert user_defined_function_handle.comment == "test_comment"

    finally:
        user_defined_function_created.drop()


def test_create_and_fetch_scala_array_input(user_defined_functions):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_scala_array_input_")

    func_body = """
        class StringHandler {
          def handleStrings(strings: Array[String]): String = {
            return concatenate(strings)
          }
          private def concatenate(strings: Array[String]): String = {
            var concatenated : String = ""
            for (newString <- strings)  {
                concatenated = concatenated + " " + newString
            }
            return concatenated
          }
        }
    """

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[Argument(name="greeting_words", datatype="ARRAY")],
            return_type=ReturnDataType(datatype="VARCHAR"),
            language_config=ScalaFunction(handler="StringHandler.handleStrings", runtime_version="2.12", packages=[]),
            body=func_body,
            comment="test_comment",
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "VARCHAR"
        assert user_defined_function_handle.language_config.runtime_version == "2.12"
        assert user_defined_function_handle.language_config.packages == []
        assert user_defined_function_handle.language_config.handler == "StringHandler.handleStrings"
        assert user_defined_function_handle.body == func_body
        assert user_defined_function_handle.comment == "test_comment"

    finally:
        user_defined_function_created.drop()


@pytest.mark.min_sf_ver("9.38.0")
def test_create_and_fetch_sql(user_defined_functions):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_sql_")

    func_body = "SELECT DECFLOAT '3.141592654'"

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[],
            return_type=ReturnDataType(datatype="DECFLOAT"),
            language_config=SQLFunction(),
            body=func_body,
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.datatype == "DECFLOAT"
        assert user_defined_function_handle.body == func_body

    finally:
        user_defined_function_created.drop()


def test_create_and_fetch_sql_hardcoded_values(user_defined_functions):
    user_defined_function_name = random_string(10, "test_create_user_defined_function_sql_")

    func_body = """
        SELECT 1, 2
        UNION ALL
        SELECT 3, 4
    """

    user_defined_function_created = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[],
            return_type=ReturnTable(
                column_list=[ColumnType(name="x", datatype="INTEGER"), ColumnType(name="y", datatype="INTEGER")]
            ),
            language_config=SQLFunction(),
            body=func_body,
        )
    )

    try:
        user_defined_function_handle = user_defined_function_created.fetch()
        assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
        assert user_defined_function_handle.return_type.column_list == [
            ColumnType(name="X", datatype="NUMBER"),
            ColumnType(name="Y", datatype="NUMBER"),
        ]
        assert user_defined_function_handle.body == func_body

    finally:
        user_defined_function_created.drop()


def test_create_and_fetch_sql_query_as_body(user_defined_functions, cursor):
    test_countries_table_name = random_string(10, "test_countries_")
    test_user_addresses_table_name = random_string(10, "test_user_addresses_")
    try:
        cursor.execute(f"""CREATE TABLE {test_countries_table_name} (
            country_code VARCHAR(10) PRIMARY KEY,
            country_name VARCHAR(255));
        """).fetchone()

        cursor.execute(f"""CREATE TABLE {test_user_addresses_table_name} (
            user_id INT,
            country_code VARCHAR(10),
            address VARCHAR(255),
            city VARCHAR(100),
            state VARCHAR(100),
            zip_code VARCHAR(20),
            PRIMARY KEY (user_id, country_code),
            FOREIGN KEY (country_code) REFERENCES {test_countries_table_name}(country_code)
        )""").fetchone()

        func_body = f"""
            SELECT DISTINCT c.country_code, c.country_name
            FROM {test_user_addresses_table_name} a, {test_countries_table_name} c
            WHERE a.user_id = id
            AND c.country_code = a.country_code
        """

        user_defined_function_name = random_string(10, "test_create_user_defined_function_sql_query_as_body_")

        user_defined_function_created = user_defined_functions.create(
            UserDefinedFunction(
                name=user_defined_function_name,
                arguments=[Argument(name="id", datatype="NUMBER")],
                return_type=ReturnTable(
                    column_list=[
                        ColumnType(name="country_code", datatype="CHAR"),
                        ColumnType(name="country_name", datatype="VARCHAR"),
                    ]
                ),
                language_config=SQLFunction(),
                body=func_body,
            )
        )

        try:
            user_defined_function_handle = user_defined_function_created.fetch()
            assert user_defined_function_handle.name.upper() == user_defined_function_name.upper()
            assert len(user_defined_function_handle.return_type.column_list) == 2
            for i in [
                ColumnType(name="country_code".upper(), datatype="VARCHAR"),
                ColumnType(name="country_name".upper(), datatype="VARCHAR"),
            ]:
                assert i in user_defined_function_handle.return_type.column_list
            assert user_defined_function_handle.body == func_body

        finally:
            user_defined_function_created.drop()

    finally:
        cursor.execute(f"DROP TABLE IF EXISTS {test_countries_table_name}").fetchone()
        cursor.execute(f"DROP TABLE IF EXISTS {test_user_addresses_table_name}").fetchone()



"""
Snowflake Function API.

The Snowflake Function API is a REST API that allows caller to create, execute and drop functions in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr

import snowflake.core.function._generated.models

from snowflake.core.function._generated.models import *  # noqa: F403
from snowflake.core.function._generated.models.function_argument import FunctionArgument


class Function(BaseModel):
    """A model object representing the Function resource.

    Constructs an object of type Function with the provided properties.

    Parameters
    __________
    name : str
        Specifies the name for the function, must be unique for the schema in which the function is created
    arguments : list[FunctionArgument]

    returns : str,  default 'TEXT'
        Specifies the type for the function return value.
    max_batch_rows : int, optional
        Specifies the max rows for batch operation.
    created_on : datetime, optional
        Date and time when the function was created.
    signature : str, optional
        Function's arguments.
    language : str, optional
        Function's language.
    body : str, optional
        Function's body.
    """

    name: StrictStr

    arguments: List[FunctionArgument]

    returns: Optional[StrictStr] = "TEXT"

    max_batch_rows: Optional[StrictInt] = None

    created_on: Optional[datetime] = None

    signature: Optional[StrictStr] = None

    language: Optional[StrictStr] = None

    body: Optional[StrictStr] = None

    __properties = [
        "function_type",
        "name",
        "arguments",
        "returns",
        "max_batch_rows",
        "created_on",
        "signature",
        "language",
        "body",
    ]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    # JSON field name that stores the object type
    __discriminator_property_name: ClassVar[str] = "function_type"

    # discriminator mappings
    __discriminator_value_class_map: ClassVar[Dict[str, str]] = {"service-function": "ServiceFunction"}

    @classmethod
    def get_discriminator_value(cls, obj: dict[str, Any]) -> Optional[str]:
        """Returns the discriminator value (object type) of the data."""
        discriminator_value = obj[cls.__discriminator_property_name]
        if discriminator_value:
            return cls.__discriminator_value_class_map.get(discriminator_value)
        else:
            return None

    __discriminator_value_to_type: ClassVar[Dict[str, str]] = {
        "ServiceFunction": "service-function",
    }

    @classmethod
    def get_child_model_discriminator_value(cls, child_model: str) -> str:
        return cls.__discriminator_value_to_type[child_model]

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Union[ServiceFunction]:
        """Create an instance of Function from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in arguments (list)
        _items = []
        if self.arguments:
            for _item in self.arguments:
                if _item:
                    _items.append(_item.to_dict())
            _dict["arguments"] = _items

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Union[ServiceFunction]:
        """Create an instance of Function from a dict."""
        # look up the object type based on discriminator mapping
        object_type = cls.get_discriminator_value(obj)
        if object_type:
            klass = getattr(snowflake.core.function._generated.models, object_type)
            return klass.from_dict(obj)
        else:
            raise ValueError(
                "Function failed to lookup discriminator value from "
                + json.dumps(obj)
                + ". Discriminator property name: "
                + cls.__discriminator_property_name
                + ", mapping: "
                + json.dumps(cls.__discriminator_value_class_map)
            )


class FunctionModel:
    def __init__(
        self,
        name: str,
        arguments: list[FunctionArgument],
        # optional properties
        returns: Optional[str] = "TEXT",
        max_batch_rows: Optional[int] = None,
        created_on: Optional[datetime] = None,
        signature: Optional[str] = None,
        language: Optional[str] = None,
        body: Optional[str] = None,
    ):
        """A model object representing the Function resource.

        Constructs an object of type Function with the provided properties.

        Parameters
        __________
        name : str
            Specifies the name for the function, must be unique for the schema in which the function is created
        arguments : list[FunctionArgument]

        returns : str,  default 'TEXT'
            Specifies the type for the function return value.
        max_batch_rows : int, optional
            Specifies the max rows for batch operation.
        created_on : datetime, optional
            Date and time when the function was created.
        signature : str, optional
            Function's arguments.
        language : str, optional
            Function's language.
        body : str, optional
            Function's body.
        """
        self.name = name
        self.arguments = arguments
        self.returns = returns
        self.max_batch_rows = max_batch_rows
        self.created_on = created_on
        self.signature = signature
        self.language = language
        self.body = body

    __properties = [
        "function_type",
        "name",
        "arguments",
        "returns",
        "max_batch_rows",
        "created_on",
        "signature",
        "language",
        "body",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Function(
            name=self.name,
            arguments=[x._to_model() for x in self.arguments] if self.arguments is not None else None,
            returns=self.returns,
            max_batch_rows=self.max_batch_rows,
            created_on=self.created_on,
            signature=self.signature,
            language=self.language,
            body=self.body,
        )

    @classmethod
    def _from_model(cls, model) -> FunctionModel:
        return model.__class__._model_class._from_model(model)

    def to_dict(self):
        """Create a dictionary of the properties from a Function.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Function object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> Union[ServiceFunctionModel]:
        """Create an instance of Function from a dict.

        This method constructs a Function object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Function
            A Function object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Function.from_dict(obj))


Function._model_class = FunctionModel



import pytest

from tests.utils import random_string

from .utils import create_service_function


pytestmark = [pytest.mark.skip_gov]


@pytest.mark.flaky
def test_iter_functions(temp_service_for_function, functions):
    funcs = []

    try:
        function_name_prefix = random_string(5, "test_func_")
        for i in range(5):
            function_name = f"{function_name_prefix}_iter_foofunc_{str(i)}"
            function_name_with_args = f"{function_name}(REAL)"
            endpoint = "ep1"

            create_service_function(
                function_name, ["REAL"], "REAL", endpoint, temp_service_for_function.name, functions
            )

            funcs.append(function_name_with_args)

        for i in range(3):
            function_name = f"{function_name_prefix}_iter_woofunc_{str(i)}"
            function_name_with_args = f"{function_name}(REAL)"
            endpoint = "end-point-2"

            create_service_function(
                function_name, ["REAL"], "REAL", endpoint, temp_service_for_function.name, functions
            )

            funcs.append(function_name_with_args)

        assert len([func.name for func in functions.iter()]) >= 8
        assert len([func.name for func in functions.iter(like=r"%%\_iter\_foofunc\_%%")]) == 5
        assert len([func.name for func in functions.iter(like=r"%%\_iter\_woofunc\_%%")]) == 3

    finally:
        for i in funcs:
            functions[i].drop()



"""
Snowflake Stream API.

The Snowflake Stream API is a REST API that you can use to access, update, and perform certain actions on Stream resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import ConfigDict, Field, StrictBool
from typing_extensions import Annotated

from snowflake.core.stream._generated.models.point_of_time import PointOfTime, PointOfTimeModel
from snowflake.core.stream._generated.models.stream_source import StreamSource


class StreamSourceView(StreamSource):
    """A model object representing the StreamSourceView resource.

    Constructs an object of type StreamSourceView with the provided properties.

    Parameters
    __________

    name : str
        Name of the source whose changes are tracked by the stream

    database_name : str, optional
        Database name to which stream source type belongs. If not provided, database  name provided in the path param will be used.

    schema_name : str, optional
        Schema name to which stream source type belongs. If not provided, schema name provided in the path param will be used.

    append_only : bool, optional
        Whether this stream is an append only stream or not
    show_initial_rows : bool, optional
        Whether this stream show initial rows on first consumption
    point_of_time : PointOfTime, optional

    base_tables : list[str], optional
        List of base tables for the stream — **Read-only:** *any user-provided value will be ignored.*
    """

    append_only: Optional[StrictBool] = None

    show_initial_rows: Optional[StrictBool] = None

    point_of_time: Optional[PointOfTime] = None

    base_tables: Optional[List[Annotated[str, Field(strict=True)]]] = None

    __properties = ["src_type", "name", "database_name", "schema_name"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> StreamSourceView:
        """Create an instance of StreamSourceView from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "base_tables",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of point_of_time
        if self.point_of_time:
            _dict["point_of_time"] = self.point_of_time.to_dict()

        _dict["src_type"] = StreamSource.get_child_model_discriminator_value("StreamSourceView")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> StreamSourceView:
        """Create an instance of StreamSourceView from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return StreamSourceView.model_validate(obj)

        _obj = StreamSourceView.model_validate(
            {
                "name": obj.get("name"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "append_only": obj.get("append_only"),
                "show_initial_rows": obj.get("show_initial_rows"),
                "point_of_time": PointOfTime.from_dict(obj.get("point_of_time"))
                if obj.get("point_of_time") is not None
                else None,
                "base_tables": obj.get("base_tables"),
            }
        )

        return _obj


class StreamSourceViewModel(StreamSource):
    def __init__(
        self,
        name: str,  # optional properties
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        append_only: Optional[bool] = None,
        show_initial_rows: Optional[bool] = None,
        point_of_time: Optional[PointOfTime] = None,
        base_tables: Optional[list[str]] = None,
    ):
        """A model object representing the StreamSourceView resource.

        Constructs an object of type StreamSourceView with the provided properties.

        Parameters
        __________

        name : str
            Name of the source whose changes are tracked by the stream

        database_name : str, optional
            Database name to which stream source type belongs. If not provided, database  name provided in the path param will be used.

        schema_name : str, optional
            Schema name to which stream source type belongs. If not provided, schema name provided in the path param will be used.

        append_only : bool, optional
            Whether this stream is an append only stream or not
        show_initial_rows : bool, optional
            Whether this stream show initial rows on first consumption
        point_of_time : PointOfTime, optional

        base_tables : list[str], optional
            List of base tables for the stream
        """
        super().__init__(
            name=name,
            database_name=database_name,
            schema_name=schema_name,
        )
        self.append_only = append_only
        self.show_initial_rows = show_initial_rows
        self.point_of_time = point_of_time
        self.base_tables = base_tables

    __properties = ["src_type", "name", "database_name", "schema_name"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return StreamSourceView(
            name=self.name,
            database_name=self.database_name,
            schema_name=self.schema_name,
            append_only=self.append_only,
            show_initial_rows=self.show_initial_rows,
            point_of_time=self.point_of_time._to_model() if self.point_of_time is not None else None,
            base_tables=self.base_tables,
        )

    @classmethod
    def _from_model(cls, model) -> StreamSourceViewModel:
        return StreamSourceViewModel(
            name=model.name,
            database_name=model.database_name,
            schema_name=model.schema_name,
            append_only=model.append_only,
            show_initial_rows=model.show_initial_rows,
            point_of_time=PointOfTimeModel._from_model(model.point_of_time) if model.point_of_time else None,
            base_tables=model.base_tables,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a StreamSourceView.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the StreamSourceView object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> StreamSourceViewModel:
        """Create an instance of StreamSourceView from a dict.

        This method constructs a StreamSourceView object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        StreamSourceView
            A StreamSourceView object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(StreamSourceView.from_dict(obj))


StreamSourceView._model_class = StreamSourceViewModel



"""Manages Snowpark Container Image Repositories.

Example:
    >>> new_image_repository = ImageRepository(name="my_imagerepo")
    >>> image_repositories = root.databases["MYDB"].schemas["MYSCHEMA"].image_repositories
    >>> my_image_repo = image_repositories.create(new_image_repository)
    >>> my_image_repo_snapshot = my_image_repo.fetch()
    >>> ir_data = image_repositories.iter(like="%my")
    >>> an_existing_repo = image_repositories["an_existing_repo"]
    >>> an_existing_repo.delete()

Refer to :class:`snowflake.core.Root` to create the ``root``.
"""

from ._generated.models.image_repository import ImageRepositoryModel as ImageRepository
from ._image_repository import ImageRepositoryCollection, ImageRepositoryResource


__all__ = ["ImageRepository", "ImageRepositoryCollection", "ImageRepositoryResource"]



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.

from .account_collection import AccountObjectCollectionParent
from .base_collection import ObjectCollection, ObjectReferenceMixin
from .database_collection import DatabaseObjectCollectionParent, DatabaseObjectReferenceMixin
from .models import (
    Clone,
    CreateMode,
    DeleteMode,
    PointOfTime,
    PointOfTimeOffset,
    PointOfTimeStatement,
    PointOfTimeTimestamp,
    TokenType,
)
from .schema_collection import SchemaObjectCollectionParent, SchemaObjectReferenceMixin


__all__ = [
    "AccountObjectCollectionParent",
    "ObjectCollection",
    "ObjectReferenceMixin",
    "DatabaseObjectCollectionParent",
    "DatabaseObjectReferenceMixin",
    "SchemaObjectCollectionParent",
    "SchemaObjectReferenceMixin",
    "Clone",
    "CreateMode",
    "DeleteMode",
    "PointOfTime",
    "PointOfTimeOffset",
    "PointOfTimeStatement",
    "PointOfTimeTimestamp",
    "TokenType",
]



def test_resume_and_suspend_recluster(dynamic_table_handle, dynamic_tables):
    dynamic_tables["dummy___table"].suspend_recluster(if_exists=True)
    dynamic_tables["dummy___table"].resume_recluster(if_exists=True)
    assert dynamic_table_handle.fetch().automatic_clustering is True
    dynamic_table_handle.suspend_recluster()
    assert dynamic_table_handle.fetch().automatic_clustering is False
    dynamic_table_handle.resume_recluster()
    assert dynamic_table_handle.fetch().automatic_clustering is True



"""
Snowflake API Integration API.

The Snowflake API Integration API is a REST API that you can use to access, update, and perform certain actions on API Integration resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import ConfigDict, StrictBool, StrictStr

from snowflake.core.api_integration._generated.models.api_hook import ApiHook


class GitHook(ApiHook):
    """A model object representing the GitHook resource.

    Constructs an object of type GitHook with the provided properties.

    Parameters
    __________
    allow_any_secret : bool,  default False
        Whether to allow any Snowflake secret when accessing the Git repository. This setting is ignored if allowed_authentication_secrets is present.
    allowed_authentication_secrets : list[str], optional
        A comma-separated list of fully-qualified Snowflake secret names that UDF or procedure handler code can use when accessing the Git repository.
    allowed_api_authentication_integrations : list[str], optional
        A comma-separated list of Snowflake security integration names that UDF or procedure handler code can use when accessing the Git repository. If empty, allow no integrations.
    """

    allow_any_secret: Optional[StrictBool] = False

    allowed_authentication_secrets: Optional[List[StrictStr]] = None

    allowed_api_authentication_integrations: Optional[List[StrictStr]] = None

    __properties = ["type"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> GitHook:
        """Create an instance of GitHook from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # set to None if allowed_authentication_secrets (nullable) is None
        if self.allowed_authentication_secrets is None:
            _dict["allowed_authentication_secrets"] = None

        # set to None if allowed_api_authentication_integrations (nullable) is None
        if self.allowed_api_authentication_integrations is None:
            _dict["allowed_api_authentication_integrations"] = None

        _dict["type"] = ApiHook.get_child_model_discriminator_value("GitHook")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> GitHook:
        """Create an instance of GitHook from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return GitHook.model_validate(obj)

        _obj = GitHook.model_validate(
            {
                "allow_any_secret": obj.get("allow_any_secret") if obj.get("allow_any_secret") is not None else False,
                "allowed_authentication_secrets": obj.get("allowed_authentication_secrets"),
                "allowed_api_authentication_integrations": obj.get("allowed_api_authentication_integrations"),
            }
        )

        return _obj


class GitHookModel(ApiHook):
    def __init__(
        self,  # optional properties
        allow_any_secret: Optional[bool] = False,
        allowed_authentication_secrets: Optional[list[str]] = None,
        allowed_api_authentication_integrations: Optional[list[str]] = None,
    ):
        """A model object representing the GitHook resource.

        Constructs an object of type GitHook with the provided properties.

        Parameters
        __________
        allow_any_secret : bool,  default False
            Whether to allow any Snowflake secret when accessing the Git repository. This setting is ignored if allowed_authentication_secrets is present.
        allowed_authentication_secrets : list[str], optional
            A comma-separated list of fully-qualified Snowflake secret names that UDF or procedure handler code can use when accessing the Git repository.
        allowed_api_authentication_integrations : list[str], optional
            A comma-separated list of Snowflake security integration names that UDF or procedure handler code can use when accessing the Git repository. If empty, allow no integrations.
        """
        super().__init__()
        self.allow_any_secret = allow_any_secret
        self.allowed_authentication_secrets = allowed_authentication_secrets
        self.allowed_api_authentication_integrations = allowed_api_authentication_integrations

    __properties = ["type"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return GitHook(
            allow_any_secret=self.allow_any_secret,
            allowed_authentication_secrets=self.allowed_authentication_secrets,
            allowed_api_authentication_integrations=self.allowed_api_authentication_integrations,
        )

    @classmethod
    def _from_model(cls, model) -> GitHookModel:
        return GitHookModel(
            allow_any_secret=model.allow_any_secret,
            allowed_authentication_secrets=model.allowed_authentication_secrets,
            allowed_api_authentication_integrations=model.allowed_api_authentication_integrations,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a GitHook.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the GitHook object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> GitHookModel:
        """Create an instance of GitHook from a dict.

        This method constructs a GitHook object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        GitHook
            A GitHook object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(GitHook.from_dict(obj))


GitHook._model_class = GitHookModel



"""
Snowflake Services API.

The Snowflake Services API is a REST API that you can use to access, update, and perform certain actions on Services resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr


class ServiceEndpoint(BaseModel):
    """A model object representing the ServiceEndpoint resource.

    Constructs an object of type ServiceEndpoint with the provided properties.

    Parameters
    __________
    name : str, optional
        User-friendly endpoint name that represents the corresponding port.
    port : int, optional
        The network port the service is listening on. NULL, when portRange is specified.
    port_range : str, optional
        The network port range the service is listening on. NULL, when port is specified.
    protocol : str,  default 'HTTP'
        Supported network protocol (TCP, HTTP, or HTTPS).
    is_public : bool,  default False
        True, if the endpoint is public, accessible from internet.
    ingress_url : str, optional
        Endpoint URL accessible from the internet — **Read-only:** *any user-provided value will be ignored.*
    """

    name: Optional[StrictStr] = None

    port: Optional[StrictInt] = None

    port_range: Optional[StrictStr] = Field(default=None, alias="portRange")

    protocol: Optional[StrictStr] = "HTTP"

    is_public: Optional[StrictBool] = False

    ingress_url: Optional[StrictStr] = None

    __properties = ["name", "port", "portRange", "protocol", "is_public", "ingress_url"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ServiceEndpoint:
        """Create an instance of ServiceEndpoint from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "ingress_url",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> ServiceEndpoint:
        """Create an instance of ServiceEndpoint from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return ServiceEndpoint.model_validate(obj)

        _obj = ServiceEndpoint.model_validate(
            {
                "name": obj.get("name"),
                "port": obj.get("port"),
                "port_range": obj.get("portRange"),
                "protocol": obj.get("protocol") if obj.get("protocol") is not None else "HTTP",
                "is_public": obj.get("is_public") if obj.get("is_public") is not None else False,
                "ingress_url": obj.get("ingress_url"),
            }
        )

        return _obj


class ServiceEndpointModel:
    def __init__(
        self,  # optional properties
        name: Optional[str] = None,
        port: Optional[int] = None,
        port_range: Optional[str] = None,
        protocol: Optional[str] = "HTTP",
        is_public: Optional[bool] = False,
        ingress_url: Optional[str] = None,
    ):
        """A model object representing the ServiceEndpoint resource.

        Constructs an object of type ServiceEndpoint with the provided properties.

        Parameters
        __________
        name : str, optional
            User-friendly endpoint name that represents the corresponding port.
        port : int, optional
            The network port the service is listening on. NULL, when portRange is specified.
        port_range : str, optional
            The network port range the service is listening on. NULL, when port is specified.
        protocol : str,  default 'HTTP'
            Supported network protocol (TCP, HTTP, or HTTPS).
        is_public : bool,  default False
            True, if the endpoint is public, accessible from internet.
        ingress_url : str, optional
            Endpoint URL accessible from the internet.
        """
        self.name = name
        self.port = port
        self.port_range = port_range
        self.protocol = protocol
        self.is_public = is_public
        self.ingress_url = ingress_url

    __properties = ["name", "port", "portRange", "protocol", "is_public", "ingress_url"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return ServiceEndpoint(
            name=self.name,
            port=self.port,
            port_range=self.port_range,
            protocol=self.protocol,
            is_public=self.is_public,
            ingress_url=self.ingress_url,
        )

    @classmethod
    def _from_model(cls, model) -> ServiceEndpointModel:
        return ServiceEndpointModel(
            name=model.name,
            port=model.port,
            port_range=model.port_range,
            protocol=model.protocol,
            is_public=model.is_public,
            ingress_url=model.ingress_url,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a ServiceEndpoint.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the ServiceEndpoint object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ServiceEndpointModel:
        """Create an instance of ServiceEndpoint from a dict.

        This method constructs a ServiceEndpoint object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        ServiceEndpoint
            A ServiceEndpoint object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(ServiceEndpoint.from_dict(obj))


ServiceEndpoint._model_class = ServiceEndpointModel



from typing import TYPE_CHECKING, Generic, Protocol, TypeVar

from snowflake.connector import SnowflakeConnection
from snowflake.core._common.base_collection import ObjectCollection, ObjectReferenceMixin


if TYPE_CHECKING:
    from snowflake.core import Root
    from snowflake.core.database import DatabaseResource
    from snowflake.core.schema import SchemaResource

T = TypeVar("T")


class SchemaObjectCollectionParent(ObjectCollection[T], Generic[T]):
    def __init__(self, schema: "SchemaResource", ref_class: type[T]) -> None:
        super().__init__(ref_class)
        self._schema = schema

    @property
    def _connection(self) -> SnowflakeConnection:
        return self.schema._connection

    @property
    def schema(self) -> "SchemaResource":
        """The SchemaResource this collection belongs to."""
        return self._schema

    @property
    def database(self) -> "DatabaseResource":
        """The DatabaseResource this collection belongs to."""
        return self.schema.collection.database

    @property
    def root(self) -> "Root":
        """The Root object this collection belongs to."""
        return self.database.collection.root

    def __repr__(self) -> str:
        type_name = type(self).__name__
        return f"<{type_name}: {self.schema.qualified_name!r}>"


class SchemaObjectReferenceProtocol(Protocol[T]):
    @property
    def collection(self) -> SchemaObjectCollectionParent[T]: ...

    @property
    def name(self) -> str: ...

    @property
    def database(self) -> "DatabaseResource": ...

    @property
    def schema(self) -> "SchemaResource": ...


class SchemaObjectReferenceMixin(Generic[T], ObjectReferenceMixin[SchemaObjectCollectionParent[T]]):
    @property
    def schema(self: SchemaObjectReferenceProtocol[T]) -> "SchemaResource":
        """The SchemaResource this reference belongs to."""
        return self.collection.schema

    @property
    def database(self: SchemaObjectReferenceProtocol[T]) -> "DatabaseResource":
        """The DatabaseResource this reference belongs to."""
        return self.collection.schema.database

    @property
    def fully_qualified_name(self: SchemaObjectReferenceProtocol[T]) -> str:
        """Return the fully qualified name of the object this reference points to."""
        return build_resource_fqn_identifier(self)

    def __repr__(self: SchemaObjectReferenceProtocol[T]) -> str:
        type_name = type(self).__name__
        return f"<{type_name}: {build_resource_fqn_identifier(self)!r}>"


def build_resource_fqn_identifier(resource: SchemaObjectReferenceProtocol[T]) -> str:
    prefix = f"{resource.database.name}.{resource.schema.name}"
    if hasattr(resource, "name_with_args"):
        return f"{prefix}.{resource.name_with_args}"
    return f"{prefix}.{resource.name}"



import pytest as pytest

from snowflake.core.exceptions import ConflictError, NotFoundError


def test_drop_and_undrop(tables, table_handle):
    table_handle.drop()
    with pytest.raises(NotFoundError):
        table_handle.fetch()

    # if exist is true, no error is raised
    table_handle.drop(if_exists=True)

    table_handle.undrop()
    assert table_handle.fetch() is not None

    with pytest.raises(ConflictError):
        table_handle.undrop()


def test_delete_deprecated(tables, table_handle):
    with pytest.warns(DeprecationWarning, match="method is deprecated; use"):
        table_handle.delete()
        with pytest.raises(NotFoundError):
            table_handle.fetch()


def test_undelete_deprecated(tables, table_handle):
    with pytest.warns(DeprecationWarning, match="method is deprecated; use"):
        table_handle.drop()
        table_handle.undelete()



import copy

from typing import Iterator

import pytest

from snowflake.core import CreateMode
from snowflake.core.exceptions import ConflictError
from snowflake.core.secret import (
    CloudProviderTokenSecret,
    GenericStringSecret,
    JwtKeyPairSecret,
    Oauth2Secret,
    PasswordSecret,
    SymmetricKeySecret,
)
from tests.integ.utils import MASKED_VALUE, random_string

from .conftest import (
    test_cloud_provider_token_secret_template,
    test_generic_string_secret_template,
    test_jwt_key_pair_secret_template,
    test_oauth2_auth_code_secret_template,
    test_oauth2_client_secret_template,
    test_password_secret_template,
    test_symmetric_key_secret_template,
)


@pytest.fixture(scope="module")
def oauth_security_integration(connection) -> Iterator[str]:
    name = random_string(10, "test_oauth_security_integration_")
    with connection.cursor() as cur:
        cur.execute(
            f"""
        CREATE SECURITY INTEGRATION {name}
          TYPE = API_AUTHENTICATION
          AUTH_TYPE = OAUTH2
          OAUTH_CLIENT_ID = 'client_id_value'
          OAUTH_CLIENT_SECRET = 'client_secret_value'
          OAUTH_ALLOWED_SCOPES = ('read', 'write')
          ENABLED = true
          COMMENT = 'Created by secret/test_create_and_fetch'
        """
        )
        try:
            yield name
        finally:
            cur.execute(f"DROP SECURITY INTEGRATION IF EXISTS {name}")


@pytest.fixture(scope="module")
def cloud_security_integration(connection) -> Iterator[str]:
    name = random_string(10, "test_cloud_security_integration_")
    with connection.cursor() as cur:
        cur.execute(
            f"""
        CREATE SECURITY INTEGRATION {name}
          TYPE = API_AUTHENTICATION
          AUTH_TYPE = AWS_IAM
          AWS_ROLE_ARN = 'arn:aws:iam::123456789012:role/SnowflakeRole'
          ENABLED = true
          COMMENT = 'Created by secret/test_create_and_fetch'
        """
        )
        try:
            yield name
        finally:
            cur.execute(f"DROP SECURITY INTEGRATION IF EXISTS {name}")


def test_create_and_fetch_cloud_provider_token_secret(secrets, cloud_security_integration: str):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_cloud_provider_token_secret_template)
        secret.name = name
        secret.api_authentication = cloud_security_integration
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, CloudProviderTokenSecret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.api_authentication == cloud_security_integration.upper()
        assert fetched_secret.comment == "Test cloud provider token secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_generic_string_secret(secrets):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_generic_string_secret_template)
        secret.name = name
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, GenericStringSecret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.secret_string == MASKED_VALUE
        assert fetched_secret.comment == "Test generic string secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_jwt_key_pair_secret(secrets):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_jwt_key_pair_secret_template)
        secret.name = name
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, JwtKeyPairSecret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.algorithm == "RSA"
        assert fetched_secret.key_length == 4096
        assert fetched_secret.comment == "Test JWT key pair secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_oauth2_client_secret(secrets, oauth_security_integration: str):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_oauth2_client_secret_template)
        secret.name = name
        secret.api_authentication = oauth_security_integration
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, Oauth2Secret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.api_authentication == oauth_security_integration.upper()
        assert fetched_secret.oauth_refresh_token is None
        assert fetched_secret.oauth_refresh_token_expiry_time is None
        assert fetched_secret.oauth_scopes == ["read", "write"]
        assert fetched_secret.comment == "Test OAuth2 client flow secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_oauth2_auth_code_secret(secrets, oauth_security_integration: str):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_oauth2_auth_code_secret_template)
        secret.name = name
        secret.api_authentication = oauth_security_integration
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, Oauth2Secret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.api_authentication == oauth_security_integration.upper()
        assert fetched_secret.oauth_refresh_token == MASKED_VALUE
        assert fetched_secret.oauth_refresh_token_expiry_time == "2030-01-01 10:00:00"
        assert fetched_secret.oauth_scopes is None
        assert fetched_secret.comment == "Test OAuth2 auth code flow secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_password_secret(secrets):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_password_secret_template)
        secret.name = name
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, PasswordSecret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.username == "snowman"
        assert fetched_secret.password == MASKED_VALUE
        assert fetched_secret.comment == "Test password secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_symmetric_key_secret(secrets):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_symmetric_key_secret_template)
        secret.name = name
        secrets.create(secret)

        fetched_secret = secret_handle.fetch()

        assert isinstance(fetched_secret, SymmetricKeySecret)
        assert fetched_secret.name == name.upper()
        assert fetched_secret.algorithm == "GENERIC"
        assert fetched_secret.comment == "Test symmetric key secret"
        assert fetched_secret.database_name == secrets.database.name
        assert fetched_secret.schema_name == secrets.schema.name
    finally:
        secret_handle.drop(if_exists=True)


def test_create_and_fetch_create_modes(secrets):
    name = random_string(10, "test_secret_create_and_fetch_")
    secret_handle = secrets[name]

    try:
        secret = copy.deepcopy(test_generic_string_secret_template)
        secret.name = name
        secret.comment = "First version"
        secrets.create(secret, mode=CreateMode.error_if_exists)
        assert secret_handle.fetch().comment == "First version"

        with pytest.raises(ConflictError):
            secrets.create(secret, mode=CreateMode.error_if_exists)

        secret.comment = "Second version"
        secrets.create(secret, mode=CreateMode.or_replace)
        assert secret_handle.fetch().comment == "Second version"

        secret.comment = "Should not change"
        secrets.create(secret, mode=CreateMode.if_not_exists)
        assert secret_handle.fetch().comment == "Second version"
    finally:
        secret_handle.drop(if_exists=True)



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.notebook import Notebook
from tests.utils import random_string


@pytest.mark.min_sf_ver("8.37.0")
def test_drop(notebooks):
    nb_name = random_string(5, "test_notebook_")
    test_nb = Notebook(name=nb_name)
    notebooks.create(test_nb)
    notebooks[test_nb.name].drop()

    with pytest.raises(NotFoundError):
        notebooks[test_nb.name].fetch()

    # creating again, making sure it's not an issue
    notebooks.create(test_nb)
    notebooks[test_nb.name].drop()



"""
Snowflake Schema API.

The Snowflake Schema API is a REST API that you can use to access, update, and perform certain actions on a Snowflake schema.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated


class Schema(BaseModel):
    """A model object representing the Schema resource.

    Constructs an object of type Schema with the provided properties.

    Parameters
    __________
    name : str
        A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
    created_on : datetime, optional
        Date and time the schema was created — **Read-only:** *any user-provided value will be ignored.*
    kind : str,  default 'PERMANENT'
        Schema type, permanent (default) or transient.
    is_default : bool, optional
        Default schema for a user — **Read-only:** *any user-provided value will be ignored.*
    is_current : bool, optional
        Current schema for the session — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database that the schema belongs to — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Name of the role that owns the schema — **Read-only:** *any user-provided value will be ignored.*
    comment : str, optional
        Optional comment in which to store information related to the schema.
    options : str, optional
        **Read-only:** *any user-provided value will be ignored.*
    managed_access : bool,  default False
        Whether this schema is a managed access schema that centralizes privilege management with the schema owner.
    retention_time : int, optional
        Number of days that historical data is retained for Time Travel — **Read-only:** *any user-provided value will be ignored.*
    dropped_on : datetime, optional
        Date and time the schema was dropped — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        Type of role that owns the object, either `ROLE` or `DATABASE_ROLE` — **Read-only:** *any user-provided value will be ignored.*
    budget : str, optional
        Budget that defines a monthly spending limit on the compute costs  for a Snowflake account or a custom group of Snowflake objects — **Read-only:** *any user-provided value will be ignored.*
    data_retention_time_in_days : int, optional
        Number of days for which Time Travel actions (CLONE and UNDROP) can be performed on the schema, as well as specifying the default Time Travel retention time for all tables created in the schema
    default_ddl_collation : str, optional
        Specifies a default collation specification for all tables added to the schema. You an override the default at the schema and individual table levels.
    log_level : str, optional
        Severity level of messages that should be ingested and made available in the active event table. Currently, Snowflake supports only `TRACE`, `DEBUG`, `INFO`, `WARN`, `ERROR`, `FATAL` and `OFF`.
    pipe_execution_paused : bool, optional
        Whether pipe execution is paused.
    max_data_extension_time_in_days : int, optional
        Maximum number of days for which Snowflake can extend the data retention period for tables in the schema to prevent streams on the tables from becoming stale.
    suspend_task_after_num_failures : int, optional
        Specifies the number of consecutive failed task runs after which the current task is suspended automatically.
    trace_level : str, optional
        How trace events are ingested into the event table. Currently, Snowflake supports only `ALWAYS`, `ON_EVENT`, and `OFF`.
    user_task_managed_initial_warehouse_size : str, optional
        Size of the compute resources to provision for the first run of the serverless task, before a task history is available for Snowflake to determine an ideal size.
    user_task_timeout_ms : int, optional
        Time limit, in milliseconds, for a single run of the task before it times out.
    serverless_task_min_statement_size : str, optional
        Specifies the minimum allowed warehouse size for the serverless task. Minimum XSMALL, Maximum XXLARGE.
    serverless_task_max_statement_size : str, optional
        Specifies the maximum allowed warehouse size for the serverless task. Minimum XSMALL, Maximum XXLARGE.
    """

    created_on: Optional[datetime] = None

    name: Annotated[str, Field(strict=True)]

    kind: Optional[StrictStr] = "PERMANENT"

    is_default: Optional[StrictBool] = None

    is_current: Optional[StrictBool] = None

    database_name: Optional[StrictStr] = None

    owner: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    options: Optional[StrictStr] = None

    managed_access: Optional[StrictBool] = False

    retention_time: Optional[StrictInt] = None

    dropped_on: Optional[datetime] = None

    owner_role_type: Optional[StrictStr] = None

    budget: Optional[StrictStr] = None

    data_retention_time_in_days: Optional[StrictInt] = None

    default_ddl_collation: Optional[StrictStr] = None

    log_level: Optional[StrictStr] = None

    pipe_execution_paused: Optional[StrictBool] = None

    max_data_extension_time_in_days: Optional[StrictInt] = None

    suspend_task_after_num_failures: Optional[StrictInt] = None

    trace_level: Optional[StrictStr] = None

    user_task_managed_initial_warehouse_size: Optional[StrictStr] = None

    user_task_timeout_ms: Optional[StrictInt] = None

    serverless_task_min_statement_size: Optional[StrictStr] = None

    serverless_task_max_statement_size: Optional[StrictStr] = None

    __properties = [
        "created_on",
        "name",
        "kind",
        "is_default",
        "is_current",
        "database_name",
        "owner",
        "comment",
        "options",
        "managed_access",
        "retention_time",
        "dropped_on",
        "owner_role_type",
        "budget",
        "data_retention_time_in_days",
        "default_ddl_collation",
        "log_level",
        "pipe_execution_paused",
        "max_data_extension_time_in_days",
        "suspend_task_after_num_failures",
        "trace_level",
        "user_task_managed_initial_warehouse_size",
        "user_task_timeout_ms",
        "serverless_task_min_statement_size",
        "serverless_task_max_statement_size",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("kind")
    def kind_validate_enum(cls, v):
        if v is None:
            return v
        if v not in ("PERMANENT", "TRANSIENT"):
            raise ValueError("must validate the enum values ('PERMANENT','TRANSIENT')")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Schema:
        """Create an instance of Schema from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "is_default",
                    "is_current",
                    "database_name",
                    "owner",
                    "options",
                    "retention_time",
                    "dropped_on",
                    "owner_role_type",
                    "budget",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # set to None if dropped_on (nullable) is None
        if self.dropped_on is None:
            _dict["dropped_on"] = None

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Schema:
        """Create an instance of Schema from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Schema.model_validate(obj)

        _obj = Schema.model_validate(
            {
                "created_on": obj.get("created_on"),
                "name": obj.get("name"),
                "kind": obj.get("kind") if obj.get("kind") is not None else "PERMANENT",
                "is_default": obj.get("is_default"),
                "is_current": obj.get("is_current"),
                "database_name": obj.get("database_name"),
                "owner": obj.get("owner"),
                "comment": obj.get("comment"),
                "options": obj.get("options"),
                "managed_access": obj.get("managed_access") if obj.get("managed_access") is not None else False,
                "retention_time": obj.get("retention_time"),
                "dropped_on": obj.get("dropped_on"),
                "owner_role_type": obj.get("owner_role_type"),
                "budget": obj.get("budget"),
                "data_retention_time_in_days": obj.get("data_retention_time_in_days"),
                "default_ddl_collation": obj.get("default_ddl_collation"),
                "log_level": obj.get("log_level"),
                "pipe_execution_paused": obj.get("pipe_execution_paused"),
                "max_data_extension_time_in_days": obj.get("max_data_extension_time_in_days"),
                "suspend_task_after_num_failures": obj.get("suspend_task_after_num_failures"),
                "trace_level": obj.get("trace_level"),
                "user_task_managed_initial_warehouse_size": obj.get("user_task_managed_initial_warehouse_size"),
                "user_task_timeout_ms": obj.get("user_task_timeout_ms"),
                "serverless_task_min_statement_size": obj.get("serverless_task_min_statement_size"),
                "serverless_task_max_statement_size": obj.get("serverless_task_max_statement_size"),
            }
        )

        return _obj


class SchemaModel:
    def __init__(
        self,
        name: str,
        # optional properties
        created_on: Optional[datetime] = None,
        kind: Optional[str] = "PERMANENT",
        is_default: Optional[bool] = None,
        is_current: Optional[bool] = None,
        database_name: Optional[str] = None,
        owner: Optional[str] = None,
        comment: Optional[str] = None,
        options: Optional[str] = None,
        managed_access: Optional[bool] = False,
        retention_time: Optional[int] = None,
        dropped_on: Optional[datetime] = None,
        owner_role_type: Optional[str] = None,
        budget: Optional[str] = None,
        data_retention_time_in_days: Optional[int] = None,
        default_ddl_collation: Optional[str] = None,
        log_level: Optional[str] = None,
        pipe_execution_paused: Optional[bool] = None,
        max_data_extension_time_in_days: Optional[int] = None,
        suspend_task_after_num_failures: Optional[int] = None,
        trace_level: Optional[str] = None,
        user_task_managed_initial_warehouse_size: Optional[str] = None,
        user_task_timeout_ms: Optional[int] = None,
        serverless_task_min_statement_size: Optional[str] = None,
        serverless_task_max_statement_size: Optional[str] = None,
    ):
        """A model object representing the Schema resource.

        Constructs an object of type Schema with the provided properties.

        Parameters
        __________
        name : str
            A Snowflake object identifier. If the identifier contains spaces or special characters, the entire string must be enclosed in double quotes. Identifiers enclosed in double quotes are also case-sensitive.
        created_on : datetime, optional
            Date and time the schema was created.
        kind : str,  default 'PERMANENT'
            Schema type, permanent (default) or transient.
        is_default : bool, optional
            Default schema for a user.
        is_current : bool, optional
            Current schema for the session.
        database_name : str, optional
            Database that the schema belongs to
        owner : str, optional
            Name of the role that owns the schema.
        comment : str, optional
            Optional comment in which to store information related to the schema.
        options : str, optional

        managed_access : bool,  default False
            Whether this schema is a managed access schema that centralizes privilege management with the schema owner.
        retention_time : int, optional
            Number of days that historical data is retained for Time Travel.
        dropped_on : datetime, optional
            Date and time the schema was dropped.
        owner_role_type : str, optional
            Type of role that owns the object, either `ROLE` or `DATABASE_ROLE`.
        budget : str, optional
            Budget that defines a monthly spending limit on the compute costs  for a Snowflake account or a custom group of Snowflake objects.
        data_retention_time_in_days : int, optional
            Number of days for which Time Travel actions (CLONE and UNDROP) can be performed on the schema, as well as specifying the default Time Travel retention time for all tables created in the schema
        default_ddl_collation : str, optional
            Specifies a default collation specification for all tables added to the schema. You an override the default at the schema and individual table levels.
        log_level : str, optional
            Severity level of messages that should be ingested and made available in the active event table. Currently, Snowflake supports only `TRACE`, `DEBUG`, `INFO`, `WARN`, `ERROR`, `FATAL` and `OFF`.
        pipe_execution_paused : bool, optional
            Whether pipe execution is paused.
        max_data_extension_time_in_days : int, optional
            Maximum number of days for which Snowflake can extend the data retention period for tables in the schema to prevent streams on the tables from becoming stale.
        suspend_task_after_num_failures : int, optional
            Specifies the number of consecutive failed task runs after which the current task is suspended automatically.
        trace_level : str, optional
            How trace events are ingested into the event table. Currently, Snowflake supports only `ALWAYS`, `ON_EVENT`, and `OFF`.
        user_task_managed_initial_warehouse_size : str, optional
            Size of the compute resources to provision for the first run of the serverless task, before a task history is available for Snowflake to determine an ideal size.
        user_task_timeout_ms : int, optional
            Time limit, in milliseconds, for a single run of the task before it times out.
        serverless_task_min_statement_size : str, optional
            Specifies the minimum allowed warehouse size for the serverless task. Minimum XSMALL, Maximum XXLARGE.
        serverless_task_max_statement_size : str, optional
            Specifies the maximum allowed warehouse size for the serverless task. Minimum XSMALL, Maximum XXLARGE.
        """
        self.created_on = created_on
        self.name = name
        self.kind = kind
        self.is_default = is_default
        self.is_current = is_current
        self.database_name = database_name
        self.owner = owner
        self.comment = comment
        self.options = options
        self.managed_access = managed_access
        self.retention_time = retention_time
        self.dropped_on = dropped_on
        self.owner_role_type = owner_role_type
        self.budget = budget
        self.data_retention_time_in_days = data_retention_time_in_days
        self.default_ddl_collation = default_ddl_collation
        self.log_level = log_level
        self.pipe_execution_paused = pipe_execution_paused
        self.max_data_extension_time_in_days = max_data_extension_time_in_days
        self.suspend_task_after_num_failures = suspend_task_after_num_failures
        self.trace_level = trace_level
        self.user_task_managed_initial_warehouse_size = user_task_managed_initial_warehouse_size
        self.user_task_timeout_ms = user_task_timeout_ms
        self.serverless_task_min_statement_size = serverless_task_min_statement_size
        self.serverless_task_max_statement_size = serverless_task_max_statement_size

    __properties = [
        "created_on",
        "name",
        "kind",
        "is_default",
        "is_current",
        "database_name",
        "owner",
        "comment",
        "options",
        "managed_access",
        "retention_time",
        "dropped_on",
        "owner_role_type",
        "budget",
        "data_retention_time_in_days",
        "default_ddl_collation",
        "log_level",
        "pipe_execution_paused",
        "max_data_extension_time_in_days",
        "suspend_task_after_num_failures",
        "trace_level",
        "user_task_managed_initial_warehouse_size",
        "user_task_timeout_ms",
        "serverless_task_min_statement_size",
        "serverless_task_max_statement_size",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Schema(
            created_on=self.created_on,
            name=self.name,
            kind=self.kind,
            is_default=self.is_default,
            is_current=self.is_current,
            database_name=self.database_name,
            owner=self.owner,
            comment=self.comment,
            options=self.options,
            managed_access=self.managed_access,
            retention_time=self.retention_time,
            dropped_on=self.dropped_on,
            owner_role_type=self.owner_role_type,
            budget=self.budget,
            data_retention_time_in_days=self.data_retention_time_in_days,
            default_ddl_collation=self.default_ddl_collation,
            log_level=self.log_level,
            pipe_execution_paused=self.pipe_execution_paused,
            max_data_extension_time_in_days=self.max_data_extension_time_in_days,
            suspend_task_after_num_failures=self.suspend_task_after_num_failures,
            trace_level=self.trace_level,
            user_task_managed_initial_warehouse_size=self.user_task_managed_initial_warehouse_size,
            user_task_timeout_ms=self.user_task_timeout_ms,
            serverless_task_min_statement_size=self.serverless_task_min_statement_size,
            serverless_task_max_statement_size=self.serverless_task_max_statement_size,
        )

    @classmethod
    def _from_model(cls, model) -> SchemaModel:
        return SchemaModel(
            created_on=model.created_on,
            name=model.name,
            kind=model.kind,
            is_default=model.is_default,
            is_current=model.is_current,
            database_name=model.database_name,
            owner=model.owner,
            comment=model.comment,
            options=model.options,
            managed_access=model.managed_access,
            retention_time=model.retention_time,
            dropped_on=model.dropped_on,
            owner_role_type=model.owner_role_type,
            budget=model.budget,
            data_retention_time_in_days=model.data_retention_time_in_days,
            default_ddl_collation=model.default_ddl_collation,
            log_level=model.log_level,
            pipe_execution_paused=model.pipe_execution_paused,
            max_data_extension_time_in_days=model.max_data_extension_time_in_days,
            suspend_task_after_num_failures=model.suspend_task_after_num_failures,
            trace_level=model.trace_level,
            user_task_managed_initial_warehouse_size=model.user_task_managed_initial_warehouse_size,
            user_task_timeout_ms=model.user_task_timeout_ms,
            serverless_task_min_statement_size=model.serverless_task_min_statement_size,
            serverless_task_max_statement_size=model.serverless_task_max_statement_size,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Schema.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Schema object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> SchemaModel:
        """Create an instance of Schema from a dict.

        This method constructs a Schema object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Schema
            A Schema object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Schema.from_dict(obj))


Schema._model_class = SchemaModel



"""
Snowflake Schema API.

The Snowflake Schema API is a REST API that you can use to access, update, and perform certain actions on a Snowflake schema.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import ConfigDict, StrictStr

from snowflake.core.schema._generated.models.point_of_time import PointOfTime


class PointOfTimeTimestamp(PointOfTime):
    """A model object representing the PointOfTimeTimestamp resource.

    Constructs an object of type PointOfTimeTimestamp with the provided properties.

    Parameters
    __________

    reference : str, optional
        Relation to the point of time. Currently, the API supports `at` and `before`.

    timestamp : str, optional
        Timestamp of the point of time (e.g. TO_TIMESTAMP(1749423600))
    """

    timestamp: Optional[StrictStr] = None

    __properties = ["point_of_time_type", "reference"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> PointOfTimeTimestamp:
        """Create an instance of PointOfTimeTimestamp from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        _dict["point_of_time_type"] = PointOfTime.get_child_model_discriminator_value("PointOfTimeTimestamp")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> PointOfTimeTimestamp:
        """Create an instance of PointOfTimeTimestamp from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return PointOfTimeTimestamp.model_validate(obj)

        _obj = PointOfTimeTimestamp.model_validate(
            {
                "reference": obj.get("reference"),
                "timestamp": obj.get("timestamp"),
            }
        )

        return _obj


class PointOfTimeTimestampModel(PointOfTime):
    def __init__(
        self,  # optional properties
        reference: Optional[str] = None,
        timestamp: Optional[str] = None,
    ):
        """A model object representing the PointOfTimeTimestamp resource.

        Constructs an object of type PointOfTimeTimestamp with the provided properties.

        Parameters
        __________

        reference : str, optional
            Relation to the point of time. Currently, the API supports `at` and `before`.

        timestamp : str, optional
            Timestamp of the point of time (e.g. TO_TIMESTAMP(1749423600))
        """
        super().__init__(
            reference=reference,
        )
        self.timestamp = timestamp

    __properties = ["point_of_time_type", "reference"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return PointOfTimeTimestamp(
            reference=self.reference,
            timestamp=self.timestamp,
        )

    @classmethod
    def _from_model(cls, model) -> PointOfTimeTimestampModel:
        return PointOfTimeTimestampModel(
            reference=model.reference,
            timestamp=model.timestamp,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a PointOfTimeTimestamp.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the PointOfTimeTimestamp object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> PointOfTimeTimestampModel:
        """Create an instance of PointOfTimeTimestamp from a dict.

        This method constructs a PointOfTimeTimestamp object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        PointOfTimeTimestamp
            A PointOfTimeTimestamp object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(PointOfTimeTimestamp.from_dict(obj))


PointOfTimeTimestamp._model_class = PointOfTimeTimestampModel
